"""
This code has been duplicated from version 1.0.0 of the snippet NetSquid-Factory, which can be found at
https://gitlab.com/softwarequtech/netsquid-snippets/netsquid-factory .
The reason it is also included here, is to allow for the creation of the figures included in the paper from data
without the need to install NetSquid-Factory.


This code can be used to evaluate the analytical results for the rate and fidelity of GHZ-state distribution
on a symmetric star-shaped network using a factory node that are presented in the paper
"Analysis of Multipartite Entanglement Distribution using a Central Quantum-Network Node"
by Guus Avis, Filip Rozpędek and Stephanie Wehner.
It includes leading-order results for the rate and fidelity, and a lower bound for the fidelity.

We note that a different convention for the depolarizing channel is used in this file than in the paper.
In the paper, a depolarizing parameter with param p has the action on one qubit
rho -> p rho + (1 - p) Id_2 / 2.
Here, the convention that is used is
rho -> (1 - p) rho + p Id_2 / 2.
This makes it easier to work to leading order in noise parameters.

Because of the difference in convention, there is the following mapping between parameter in the paper and parameters
used here:
link_depolar_prob = 1 - p_link
bsm_depolar_prob = 1 - p_BSM
mem_depolar_prob = 1 - p_mem

Additionally, we here don't use the parameter p_GHZ for the noise in the GHZ state as it is locally prepared.
Instead, we use the fidelity of the prepared state,
ghz_fidelity = 1 - (2 ** num_qubits - 1) / 2 ** num_qubits * p_ghz
(see also documentation of the function `fidelity_to_depolarizing_prob()`).


"""
import itertools


def factory_waiting_time(num_remote_nodes, link_success_prob, bsm_success_prob, **kwargs):
    """Leading-order expression of the average waiting time to deliver a single GHZ state with the GHZ-factory protocol.

    The expression is to leading order in the link success probability.
    It thus becomes exact in the limit when the link success probability goes to zero.

    Parameters
    ----------
    num_remote_nodes : int
        Number of remote nodes that attempt to share a GHZ state through the GHZ-factory protocol.
    link_success_prob : float
        Probability per attempt that entanglement distribution between a remote node and the factory succeeds.
    bsm_success_prob : float
        Success probability of Bell-state measurements.

    """

    num_remote_nodes = int(num_remote_nodes)  # in case it's passed as a float
    # find harmonic number H_N where N = num_remote_nodes
    harmonic_number = 1
    for i in range(2, num_remote_nodes + 1):
        harmonic_number += 1 / i

    return harmonic_number / link_success_prob / bsm_success_prob ** num_remote_nodes


def factory_fidelity(num_remote_nodes, link_success_prob, mem_depolar_prob, link_depolar_prob,
                     bsm_depolar_prob, ghz_fidelity=1, bound=False, **kwargs):
    """Analytical results for the fidelity achieved with the GHZ-factory protocol (leading-order expression or bound).

    Parameter `bound` can be used to toggle between a strict lower bound on the fidelity and a leading-order expression.

    The leading-order expression is to leading order in the link success probability (`link_success_prob`)
    and the probability that a qubit undergoes a depolarizing error in memory while performing a single attempt
    at Bell-state distribution (`mem_depolar_prob`).

    Parameters
    ----------
    num_remote_nodes : int
        Number of remote nodes that attempt to share a GHZ state through the GHZ-factory protocol.
    link_success_prob : float
        Probability per attempt that entanglement distribution between a remote node and the factory succeeds.
    ghz_fidelity : float
        Fidelity of the GHZ state that is created at the GHZ factory.
    mem_depolar_prob : float
        Depolarizing noise in quantum memory per time step for both the GHZ factory and end nodes.
    link_depolar_prob : float
        Depolarizing noise on EPR pairs delivered at entanglement generation.
        The state that is delivered is assumed to be (1 - link_depolar_prob)|Phi+><Phi+| + link_depolar_prob * Id / 2
    bsm_depolar_prob : float
        Depolarizing probability on each qubit participating in a Bell-state measurement.
    bound : bool (optional)
        If True, a lower bound is calculated.
        If False, a leading-order expression is calculated.

    """
    g_func = _g_function if not bound else _g_function_lower_bound

    num_remote_nodes = int(num_remote_nodes)  # in case it's passed as a float
    ghz_depolarizing_prob = fidelity_to_depolarizing_prob(num_qubits=num_remote_nodes, fidelity=ghz_fidelity)
    link_bsm_depolar = (1 - link_depolar_prob) * (1 - bsm_depolar_prob) ** 2
    prob = 2 * mem_depolar_prob - mem_depolar_prob ** 2
    set_all_integers = set(range(1, num_remote_nodes + 1))

    fidelity = 0
    for i in range(num_remote_nodes + 1):
        i_even = i % 2 == 0
        if not i_even and i != num_remote_nodes:
            continue
        prefactor = (1 / 2) ** num_remote_nodes if i_even else 0
        if i == num_remote_nodes:
            prefactor += 1 / 2
        subsets_length_i = itertools.combinations(set_all_integers, i)
        gs = 0
        for subset in subsets_length_i:
            g = g_func(num_remote_nodes=num_remote_nodes,
                       link_success_prob=link_success_prob,
                       subset=set(subset),
                       prob=prob)
            gs += g
        fidelity += prefactor * link_bsm_depolar ** i * gs

    fidelity *= 1 - ghz_depolarizing_prob

    fidelity += ghz_depolarizing_prob / 2 ** num_remote_nodes

    return fidelity


def _g_function(num_remote_nodes, link_success_prob, subset, prob):
    """Leading-order expression for the function G in Appendix D of the paper.

    For definition and derivation, see Appendix D of the paper
    "Analysis of Multipartite Entanglement Distribution using a Central Quantum-Network Node"
    Guus Avis, Filip Rozpędek and Stephanie Wehner.

    Parameters
    ----------
    num_remote_nodes : int
        Number of remote nodes that attempt to share a GHZ state through the GHZ-factory protocol.
    link_success_prob : float
        Probability per attempt that entanglement distribution between a remote node and the factory succeeds.
    subset : set of int
        A subset of {1, 2, ..., num_remote_nodes}.
    prob : float
        Number between 0 and 1, typically a depolarizing probability.

    """
    _check_set(subset, num_remote_nodes)
    value = 1
    for i in range(1, num_remote_nodes + 1):
        value *= (num_remote_nodes + 1 - i) * link_success_prob
        value /= sum([1 for j in subset if j < i]) * prob + (num_remote_nodes + 1 - i) * link_success_prob
    return value


def _g_function_lower_bound(num_remote_nodes, link_success_prob, subset, prob):
    """Lower bound for the function G in Appendix D of the paper

    For definition and derivation, see Appendix D of the paper
    "Analysis of Multipartite Entanglement Distribution using a Central Quantum-Network Node"
    Guus Avis, Filip Rozpędek and Stephanie Wehner.

    Parameters
    ----------
    num_remote_nodes : int
        Number of remote nodes that attempt to share a GHZ state through the GHZ-factory protocol.
    link_success_prob : float
        Probability per attempt that entanglement distribution between a remote node and the factory succeeds.
    subset : set of int
        A subset of {1, 2, ..., num_remote_nodes}.
    prob : float
        Number between 0 and 1, typically a depolarizing probability.

    """
    _check_set(subset, num_remote_nodes)
    value = 1
    for i in range(1, num_remote_nodes + 1):
        num_smaller_cs = sum([1 for j in subset if j < i])
        one_min_p = (1 - prob) ** num_smaller_cs
        value *= (num_remote_nodes + 1 - i) * link_success_prob
        value *= (1 - link_success_prob) ** (num_remote_nodes - i) * one_min_p
        value /= 1 - (1 - link_success_prob) ** (num_remote_nodes + 1 - i) * one_min_p
    return value


def fidelity_to_depolarizing_prob(fidelity, num_qubits):
    """Depolarizing probability needed to reach a certain fidelity by depolarizing a pure state.

    Depolarizing channel: rho -> rho' = (1 - depolarizing_prob) rho + depolarizing_prob * Id / 2 ** num_qubits.
    This function returns depolarizing_prob needed for rho' to have a specific fidelity to rho in case rho is pure.
    The fidelity of rho to rho is 1, and the fidelity of rho to Id is also 1 (because rho is pure).
    Therefore, the fidelity of rho' to rho is (1 - depolarizing_prob) + depolarizing_prob / 2 ** num_qubits.
    This can be rewritten as fidelity = 1 - (2 ** num_qubits - 1) / 2 ** num_qubits * depolarizing_prob.
    This can be inverted to give depolarizing_probability = 2 ** num_qubits / (2 ** num_qubits - 1) * (1 - fidelity).

    Parameters
    ----------
    fidelity : float
        Fidelity (squared definition) of the depolarized state to the state before the depolarizing channel.
    num_qubits : float
        Number of qubits in the state that undergoes a depolarizing channel.

    Returns
    -------
    depolarizing_prob : float
        Fidelity of the depolarized state to the state before depolarization.

    """
    return 2 ** num_qubits / (2 ** num_qubits - 1) * (1 - fidelity)


def depolarizing_prob_to_fidelity(depolarizing_prob, num_qubits):
    """Fidelity of a depolarized pure state to that pure state.

    Depolarizing channel: rho -> rho' = (1 - depolarizing_prob) rho + depolarizing_prob * Id / 2 ** num_qubits.
    This function returns the fidelity (squared definition) of rho' to rho in case rho is a pure state.
    The fidelity of rho to rho is 1, and the fidelity of rho to Id is also 1 (because rho is pure).
    Therefore, the fidelity of rho to rho' is (1 - depolarizing_prob) + depolarizing_prob / 2 ** num_qubits.
    This can be rewritten as fidelity = 1 - (2 ** num_qubits - 1) / 2 ** num_qubits * depolarizing_prob.

    Parameters
    ----------
    depolarizing_prob : float
        Probability that state becomes maximally mixed by the depolarizing channel.
    num_qubits : float
        Number of qubits in the state that undergoes a depolarizing channel.

    Returns
    -------
    fidelity : float
        Fidelity of the depolarized state to the state before depolarization.

    """
    return 1 - (2 ** num_qubits - 1) / 2 ** num_qubits * depolarizing_prob


def _check_set(subset, num_remote_nodes):
    """Check if subset is a proper subset of {1, ..., num_remote_nodes}."""
    if not isinstance(subset, set):
        raise TypeError("subset must be a set.")
    if not subset.issubset(set(range(1, num_remote_nodes + 1))):
        raise ValueError("subset must be a subset of {1, 2, ..., num_remote_nodes}.")
